from promptmage.storage.data_store import DataStore

from nicegui import ui


def create_runs_view(data_store: DataStore):

    def build_ui():
        with ui.column():
            ui.label("Runs")
            runs = data_store.get_all_data()
            for run_id, run_data in runs.items():
                with ui.card():
                    with ui.expansion(
                        f"Run {run_id} of step \"{run_data['step_name']}\""
                    ).classes("w-full"):
                        ui.label(f"Run time: {run_data['run_time']}")
                        ui.label(f"Prompt: {run_data['prompt']}")
                        ui.label(f"Input data: {run_data['input_data']}")
                        ui.label(f"Output data: {run_data['output_data']}")
                    ui.update()

    return build_ui



"""This module contains the RunData class, which is used to represent the data for a single run of a promptmage flow."""

import uuid
from datetime import datetime
from typing import Dict

from promptmage.prompt import Prompt


class RunData:
    """A class that represents the data for a single run of a promptmage flow."""

    def __init__(
        self,
        step_name: str,
        prompt: Prompt,
        input_data: Dict,
        output_data: Dict,
        run_id: str | None = None,
        run_time: datetime = datetime.now(),
    ):
        self.run_id = run_id if run_id else str(uuid.uuid4())
        self.step_name = step_name
        self.run_time = run_time
        self.prompt = prompt
        self.input_data = input_data
        self.output_data = output_data

    def __repr__(self) -> str:
        return (
            f"RunData(run_id={self.run_id}, "
            f"step_name={self.step_name}, "
            f"run_time={self.run_time}, "
            f"prompt={self.prompt}, "
            f"input_data={self.input_data}, "
            f"output_data={self.output_data})"
        )

    def to_dict(self) -> Dict:
        return {
            "prompt": self.prompt.to_dict(),
            "step_name": self.step_name,
            "input_data": self.input_data,
            "output_data": self.output_data,
            "run_id": self.run_id,
            "run_time": self.run_time,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            data["step_name"],
            data["prompt"],
            data["input_data"],
            data["output_data"],
            data["run_id"],
            data["run_time"],
        )



from contextlib import contextmanager
from nicegui import ui

from .menu import menu


@contextmanager
def frame(navigation_title: str):
    """Custom page frame to share the same styling and behavior across all pages"""
    ui.colors(
        primary="#6E93D6", secondary="#53B689", accent="#111B1E", positive="#53B689"
    )
    with ui.header():
        ui.avatar("img:/static/promptmage-logo.png", size=128)
        ui.space()
        ui.label(navigation_title)
        ui.space()
        with ui.row():
            menu()
    with ui.column():  # .classes("absolute-center items-center"):
        yield



from nicegui import ui


def menu() -> None:
    ui.link("Home", "/").classes(replace="text-white")
    ui.link("Runs", "/runs").classes(replace="text-white")
    ui.link("Prompts", "/prompts").classes(replace="text-white")



import uuid
import inspect
from loguru import logger
from functools import wraps
from collections import defaultdict, deque
from typing import Dict, Callable, List


# Local imports
from .prompt import Prompt
from .run_data import RunData
from .storage import PromptStore, DataStore


class PromptMage:
    """A class to represent a PromptMage instance.

    Attributes:
        name (str): The name of the PromptMage instance.
        prompt_store (PromptStore): The prompt store to use for storing prompts.
        data_store (DataStore): The data store to use for storing data.
        steps (Dict): A dictionary of steps in the PromptMage instance.
    """

    def __init__(
        self,
        name: str,
        prompt_store: PromptStore | None = None,
        data_store: DataStore | None = None,
    ):
        self.name: str = name
        self.prompt_store = prompt_store
        self.data_store = data_store
        self.steps: Dict = {}
        logger.info(f"Initialized PromptMage with name: {name}")

        self.dependencies = defaultdict(list)
        self.graph = defaultdict(list)
        self.indegree = defaultdict(int)

    def step(
        self, name: str, prompt_name: str, depends_on: List[str] | str | None = None
    ) -> Callable:
        """Decorator to add a step to the PromptMage instance.

        Args:
            name (str): The name of the step.
            prompt_name (str): The name of the prompt to use for this step. Defaults to None.
            depends_on (str, optional):
        """

        def decorator(func):

            @wraps(func)
            def wrapper(*args, **kwargs):
                logger.info(f"Running step: {name}...")
                logger.info(f"Step input: {args}, {kwargs}")
                # Get the prompt from the backend if it exists.
                prompt = self.prompt_store.get_prompt(prompt_name)
                # extract the template variables from the function signature
                if prompt.template_vars == []:
                    sig = inspect.signature(func)
                    prompt.template_vars = [
                        param for param in sig.parameters if param != "prompt"
                    ]
                    # Store the updated prompt
                    self.prompt_store.store_prompt(prompt)
                # Add the prompt to the kwargs
                kwargs["prompt"] = prompt
                try:
                    results = func(*args, **kwargs)
                except KeyError as e:
                    results = f"Error: {e}"

                # Store input and output data
                if self.data_store:
                    run_data = RunData(
                        step_name=name,
                        prompt=prompt,
                        input_data={k: v for k, v in kwargs.items() if k != "prompt"},
                        output_data=results,
                    )
                    self.data_store.store_data(run_data)
                logger.info(f"Step output: {results}")
                return results

            self.steps[name] = MageStep(
                name=name,
                func=wrapper,
                prompt_store=self.prompt_store,
                prompt_name=prompt_name,
                depends_on=depends_on,
            )

            # store the dependencies
            if depends_on:
                self.dependencies[name] = (
                    depends_on if isinstance(depends_on, list) else [depends_on]
                )
            else:
                self.dependencies[name] = []

            return wrapper

        return decorator

    def _build_dependency_graph(self):
        # Build the graph and compute in-degrees of each node
        self.graph.clear()
        self.indegree.clear()

        for func_id, deps in self.dependencies.items():
            for dep in deps:
                self.graph[dep].append(func_id)
                self.indegree[func_id] += 1
            if func_id not in self.indegree:
                self.indegree[func_id] = 0

    def topological_sort(self):
        order = []
        queue = deque([node for node in self.indegree if self.indegree[node] == 0])

        while queue:
            node = queue.popleft()
            order.append(node)
            for neighbor in self.graph[node]:
                self.indegree[neighbor] -= 1
                if self.indegree[neighbor] == 0:
                    queue.append(neighbor)

        if len(order) == len(self.indegree):
            return order
        else:
            raise ValueError(
                "Cyclic dependency detected. This is currently not supported."
            )

    def get_run_function(self, start_from=None) -> Callable:
        """Returns a function that runs the PromptMage graph starting from the given step.

        It has the same signature as the first function in the graph and takes these inputs.
        """
        self._build_dependency_graph()
        order = self.topological_sort()

        start_index = 0
        if start_from:
            start_index = order.index(start_from)

        def run_function(**initial_inputs):
            results = {}
            for func_id in order[start_index:]:
                func_node: MageStep = self.steps[func_id]
                if func_id == order[start_index] and initial_inputs:
                    # If it's the first function and initial_inputs are provided, use them
                    results[func_id] = func_node.execute(**initial_inputs)
                else:
                    if self.dependencies[func_id]:
                        inputs = {
                            param: self.steps[dep].result
                            for dep, param in zip(
                                self.dependencies[func_id],
                                func_node.signature.parameters,
                            )
                        }
                    else:
                        inputs = {
                            param: func_node.input_values[param]
                            for param in func_node.signature.parameters
                        }
                    results[func_id] = func_node.execute(**inputs)
            # return the results of the last function
            return results[order[-1]]

        # Set the signature of the returned function to match the first function in the graph
        first_func_id = order[start_index]
        first_func_node: MageStep = self.steps[first_func_id]
        run_function.__signature__ = first_func_node.signature

        return run_function

    def __repr__(self) -> str:
        return f"PromptMage(name={self.name}, steps={list(self.steps.keys())})"


class MageStep:
    def __init__(
        self,
        name: str,
        func: Callable,
        prompt_store: PromptStore,
        prompt_name: str | None = None,
        depends_on: str | None = None,
    ):
        self.step_id = str(uuid.uuid4())
        self.name = name
        self.func = func
        self.signature = inspect.signature(func)
        self.prompt_store = prompt_store
        self.prompt_name = prompt_name
        self.depends_on = depends_on

        # store inputs and results
        self.input_values = {}
        self.result = None

        # Initialize input values with default parameter values
        for param in self.signature.parameters.values():
            if param.name == "prompt":
                continue
            if param.default is not inspect.Parameter.empty:
                self.input_values[param.name] = param.default
            else:
                self.input_values[param.name] = None

        # callbacks for the frontend
        self._input_callbacks = []
        self._output_callbacks = []

    def execute(self, **inputs):
        for key, value in inputs.items():
            self.input_values[key] = value
        for callback in self._input_callbacks:
            callback()
        self.result = self.func(**self.input_values)
        # run the callbacks
        for callback in self._output_callbacks:
            callback()
        return self.result

    def __repr__(self):
        return f"Step(step_id={self.step_id}, name={self.name}, prompt_name={self.prompt_name}, depends_on={self.depends_on})"

    def get_prompt(self) -> Prompt:
        return self.prompt_store.get_prompt(self.prompt_name)

    def set_prompt(self, prompt: Prompt):
        self.prompt_store.store_prompt(prompt)

    def on_input_change(self, callback):
        self._input_callbacks.append(callback)

    def on_output_change(self, callback):
        self._output_callbacks.append(callback)



from promptmage.storage.prompt_store import PromptStore

from nicegui import ui


def create_prompts_view(prompt_store: PromptStore):

    def build_ui():
        with ui.column():
            ui.label("Prompts")
            prompts = prompt_store.get_prompts()
            prompts_dict = {prompt.id: prompt.to_dict() for prompt in prompts}
            for prompt_id, prompt in prompts_dict.items():
                with ui.card():
                    with ui.expansion(
                        f"Prompt {prompt_id} with name \"{prompt['name']}\""
                    ).classes("w-full"):
                        ui.label(f"Name: {prompt['name']}")
                        ui.label(f"ID: {prompt['id']}")
                        ui.label(f"System: {prompt['system']}")
                        ui.label(f"User: {prompt['user']}")
                        ui.label(f"Version: {prompt['version']}")
                        ui.label(f"Template Variables: {prompt['template_vars']}")
                        with ui.row():
                            ui.button(
                                "Use Prompt",
                                on_click=lambda prompt_id=prompt_id: prompt_store.use_prompt(
                                    prompt_id
                                ),
                            )
                            ui.button(
                                "Edit Prompt",
                                on_click=lambda prompt_id=prompt_id: prompt_store.edit_prompt(
                                    prompt_id
                                ),
                            )
                            ui.button(
                                "Delete Prompt",
                                on_click=lambda prompt_id=prompt_id: prompt_store.delete_prompt(
                                    prompt_id
                                ),
                            )
                    ui.update()

    return build_ui



"""This ui element represent the input, prompt and output of a callable step in the PromptMage."""

import inspect
from nicegui import ui, run
from loguru import logger

from promptmage import PromptMage


def create_main_runner(mage: PromptMage):
    input_fields = {}
    result_field = None
    flow_func = mage.get_run_function(start_from=None)

    async def run_function():
        inputs = {name: field.value for name, field in input_fields.items()}
        result = await run.io_bound(flow_func, **inputs)
        result_field.set_content(f"{str(result)}")
        result_field.update()

    def build_ui():
        nonlocal result_field
        with ui.column().style(
            "border: 1px solid #ddd; padding: 20px; border-radius: 10px; box-shadow: 0 4px 8px rgba(0, 0, 0, 0.1); margin-bottom: 20px; width: 600px; height: 800px; overflow-y: auto;"
        ):
            ui.label("Flow Runner").style("font-weight: bold; font-size: 1.5em;")
            ui.label("Inputs:").style("margin-top: 20px; font-weight: bold;")
            for param in inspect.signature(flow_func).parameters.values():
                if param.name != "prompt":
                    with ui.row():
                        ui.label(f"{param.name}:").style("width: 100px;")
                        input_fields[param.name] = ui.textarea().style(
                            "flex-grow: 2; overflow: auto;"
                        )

            ui.button("Run", on_click=run_function).style("margin-top: 10px;")
            ui.separator()
            ui.label("Result:").style("margin-top: 20px; font-weight: bold;")
            result_field = ui.markdown("").style(
                "margin-top: 20px; color: blue; height: 200px; overflow-y: auto;"
            )

    return build_ui



"""This module contains the SQLiteBackend class, which is a subclass of the StorageBackend class. It is used to store the data in a SQLite database."""

import sqlite3
import json
from loguru import logger
from typing import List, Dict

from promptmage.prompt import Prompt
from promptmage.run_data import RunData
from promptmage.storage.storage_backend import StorageBackend


class SQLitePromptBackend(StorageBackend):
    """A class that stores the prompts in a SQLite database."""

    def __init__(self, db_path: str = "promptmage.db"):
        self.db_path = db_path
        self._init_db()
        self._create_table()

    def _init_db(self):
        conn = None
        try:
            conn = sqlite3.connect(self.db_path)
        except sqlite3.Error as e:
            logger.error(f"Error connecting to SQLite database: {e}")
        finally:
            if conn:
                conn.close()

    def _create_table(self):
        create_table_sql = """
        CREATE TABLE IF NOT EXISTS prompts (
            id TEXT PRIMARY KEY,
            name TEXT NOT NULL,
            system TEXT NOT NULL,
            user TEXT NOT NULL,
            version INTEGER NOT NULL,
            template_vars TEXT NOT NULL
        );
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(create_table_sql)
            conn.commit()

    def store_prompt(self, prompt: Prompt):
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(
                "INSERT INTO prompts VALUES (?, ?, ?, ?, ?, ?)",
                (
                    prompt.id,
                    prompt.name,
                    prompt.system,
                    prompt.user,
                    prompt.version,
                    ",".join(prompt.template_vars),
                ),
            )
            conn.commit()

    def get_prompt(self, prompt_name: str) -> Prompt:
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM prompts WHERE name=?", (prompt_name,))
            row = cursor.fetchone()
            if row is None:
                return None
            return Prompt(
                id=row[0],
                name=row[1],
                system=row[2],
                user=row[3],
                version=row[4],
                template_vars=row[5].split(","),
            )

    def get_prompts(self) -> List[Prompt]:
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM prompts")
            rows = cursor.fetchall()
            prompts = []
            for row in rows:
                prompts.append(
                    Prompt(
                        id=row[0],
                        name=row[1],
                        system=row[2],
                        user=row[3],
                        version=row[4],
                        template_vars=row[5].split(","),
                    )
                )
            return prompts


class SQLiteDataBackend(StorageBackend):
    """A class that stores the data in a SQLite database."""

    def __init__(self, db_path: str = "promptmage.db"):
        self.db_path = db_path
        self._init_db()
        self._create_table()

    def _init_db(self):
        conn = None
        try:
            conn = sqlite3.connect(self.db_path)
        except sqlite3.Error as e:
            logger.error(f"Error connecting to SQLite database: {e}")
        finally:
            if conn:
                conn.close()

    def _create_table(self):
        create_table_sql = """
        CREATE TABLE IF NOT EXISTS data (
            run_id TEXT PRIMARY KEY,
            run_time TEXT NOT NULL,
            step_name TEXT NOT NULL,
            prompt TEXT NOT NULL,
            input_data TEXT,
            output_data TEXT
        );
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(create_table_sql)
            conn.commit()

    def store_data(self, run_data: RunData):
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(
                "INSERT INTO data VALUES (?, ?, ?, ?, ?, ?)",
                (
                    run_data.run_id,
                    run_data.run_time,
                    run_data.step_name,
                    json.dumps(run_data.prompt.to_dict()),
                    json.dumps(run_data.input_data),
                    json.dumps(run_data.output_data),
                ),
            )
            conn.commit()

    def get_data(self, run_id: str) -> RunData:
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM data WHERE run_id=?", (run_id,))
            row = cursor.fetchone()
            if row is None:
                return None
            return RunData(
                run_id=row[0],
                run_time=row[1],
                step_name=row[2],
                prompt=Prompt.from_dict(json.loads(row[3])),
                input_data=json.loads(row[4]),
                output_data=json.loads(row[5]),
            )

    def get_all_data(self) -> Dict[str, RunData]:
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM data")
            rows = cursor.fetchall()
            run_data = {}
            for row in rows:
                run_data[row[0]] = RunData(
                    run_id=row[0],
                    run_time=row[1],
                    step_name=row[2],
                    prompt=Prompt.from_dict(json.loads(row[3])),
                    input_data=json.loads(row[4]),
                    output_data=json.loads(row[5]),
                ).to_dict()

            return run_data



"""This module contains the command line interface for the PromptMage package."""

import click
import uvicorn

from promptmage.utils import get_flow
from promptmage.api import PromptMageAPI
from promptmage.frontend import PromptMageFrontend


@click.command()
@click.argument("file_path", type=click.Path(exists=True))
@click.option(
    "--host", default="0.0.0.0", help="The host IP to run the FastAPI server on."
)
@click.option(
    "--port", default=8000, type=int, help="The port to run the FastAPI server on."
)
def serve(file_path, host, port):
    """Serve a FastAPI application containing a FlowForge instance from the given file."""
    current_flow = get_flow(file_path)

    if not current_flow:
        raise ValueError("No PromptMage instance found in the module.")

    # create the FastAPI app
    app = PromptMageAPI(mage=current_flow).get_app()

    # create the frontend app
    frontend = PromptMageFrontend(mage=current_flow)
    frontend.init_from_api(app)

    # Run the applications
    uvicorn.run(app, host=host, port=port)


if __name__ == "__main__":
    serve()



"""This module contains the API for the PromptMage package."""

import inspect
import pkg_resources
from pathlib import Path

from fastapi import FastAPI, Path, Query
from fastapi.staticfiles import StaticFiles
from fastapi.responses import HTMLResponse
from fastapi.middleware.cors import CORSMiddleware


from promptmage import PromptMage


class PromptMageAPI:
    """A class that creates a FastAPI application to serve a PromptMage instance."""

    def __init__(self, mage: PromptMage):
        self.mage = mage

    def get_app(self) -> FastAPI:
        """Create a FastAPI application to serve the PromptMage instance."""
        app = FastAPI(
            title=f"PromptMage API: {self.mage.name}", description="API for PromptMage."
        )

        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

        # create index endpoint
        @app.get("/")
        async def index():
            return HTMLResponse(
                """<h1>Welcome to the PromptMage</h1>"""
                """<h2>To see the API docs 📖, go to <a href='/docs'>/docs</a></h2>"""
                """<h2>To access the API 🔮, go to <a href='/api'>/api</a></h2>"""
                """<h2>To access the GUI 🧙, go to <a href='/gui'>/gui</a></h2>"""
            )

        @app.get("/api")
        async def root():
            return HTMLResponse("<h1>Welcome to the PromptMage API</h1>")

        @app.get("/api/prompts")
        async def list_prompts():
            return self.mage.prompt_store.get_prompts()

        @app.get("/api/data")
        async def list_data():
            return self.mage.data_store.get_all_data()

        # create the endpoints for each step
        step_list = []
        for step_name, step in self.mage.steps.items():
            signature = inspect.signature(step.func)
            path = f"/api/{step_name}"
            params, path_variables = self._parameters_from_signature(signature)
            path += path_variables

            # Update the signature for the endpoint function
            new_signature = signature.replace(parameters=params)
            endpoint_func = self.create_endpoint_function(step.func, new_signature)
            setattr(
                endpoint_func, "__signature__", new_signature
            )  # Update the signature for FastAPI to recognize

            # Add the route to FastAPI
            app.add_api_route(path, endpoint_func, methods=["GET"])
            step_list.append({"name": step_name, "path": path})

        # create an endpoint to run the full dependency graph of the flow
        run_function = self.mage.get_run_function()
        signature = inspect.signature(run_function)
        path = "/api/run_flow"
        params, path_variables = self._parameters_from_signature(signature)
        path += path_variables
        new_signature = signature.replace(parameters=params)
        endpoint_func = self.create_endpoint_function(run_function, new_signature)
        setattr(endpoint_func, "__signature__", new_signature)
        app.add_api_route(path, endpoint_func, methods=["GET"])

        # add a route to list all available steps with their names and input variables
        @app.get("/api/steps")
        async def list_steps():
            return step_list

        static_files_path = pkg_resources.resource_filename("promptmage", "static/")
        app.mount(
            "/static/",
            StaticFiles(directory=static_files_path, html=True),
            name="static",
        )
        return app

    def _parameters_from_signature(self, signature):
        params = []
        path = ""
        for name, param in signature.parameters.items():
            # ignore prompt parameter
            if name == "prompt":
                continue
            if param.default is inspect.Parameter.empty:
                # Assume required parameters are path parameters
                new_param = inspect.Parameter(
                    name,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=Path(..., description=f"Path parameter `{name}`"),
                    annotation=param.annotation,
                )
                path += f"/{{{name}}}"  # Add to the path
            else:
                # Parameters with defaults are query parameters
                new_param = inspect.Parameter(
                    name,
                    kind=inspect.Parameter.KEYWORD_ONLY,
                    default=Query(
                        param.default, description=f"Query parameter `{name}`"
                    ),
                    annotation=param.annotation,
                )
            params.append(new_param)
        return params, path

    def create_endpoint_function(self, func, params):
        # Define the endpoint function using dynamic parameters
        async def endpoint(*args, **kwargs):
            # Directly pass the keyword arguments to the function
            return func(*args, **kwargs)

        return endpoint
